from ._libpkgconf import ffi, lib

from .constants import Flags
from .logger import logger


@ffi.def_extern()
def error_handler(msg: str, client, data) -> bool:
    """ This is a Python callback to handle errors """
    log_level = ffi.from_handle(data)

    logger.log(log_level, ffi.string(msg).decode(errors='replace').rstrip())
    return True


@ffi.def_extern()
def print_list_entry(entry, data) -> bool:
    results = ffi.from_handle(data)

    if entry.flags & lib.PKGCONF_PKG_PROPF_UNINSTALLED:
        return False
    
    entry_id = ffi.string(entry.id).decode()
    realname = ffi.string(entry.realname).decode()
    description = ffi.string(entry.description).decode()
    results.append(f'{entry_id:<30s} {realname} - {description}')

    return False


@ffi.def_extern()
def print_package_entry(entry, data) -> bool:
    results = ffi.from_handle(data)

    if entry.flags & lib.PKGCONF_PKG_PROPF_UNINSTALLED:
        return False
    
    results.append(str(entry.id))

    return False


@ffi.def_extern()
def print_license(client, pkg, data) -> None:
    result = ffi.from_handle(data)

    if pkg.flags & lib.PKGCONF_PKG_PROPF_VIRTUAL:
        return
    
    pkg_id = ffi.string(pkg.id).decode()
    # NOASSERTION is the default when the license is unknown, per SPDX spec § 3.15
    pkg_license = ffi.string(pkg.license).decode() if pkg.license != ffi.NULL else "NOASSERTION"
    result.append(f'{pkg_id}: {pkg_license}')


@ffi.def_extern()
def check_uninstalled(client, pkg, data) -> None:
    ret = ffi.from_handle(data)

    if pkg.flags & lib.PKGCONF_PKG_PROPF_UNINSTALLED:
        ret[0] = True


@ffi.def_extern()
def filter_cflags(client, frag, data) -> bool:
    want_flags, want_fragment_filter = ffi.from_handle(data)

    if Flags.KEEP_SYSTEM_CFLAGS not in want_flags and lib.pkgconf_fragment_has_system_dir(client, frag):
        return False
    
    frag_type = frag.type.decode()
    if (want_fragment_filter is not None
            and (frag_type not in want_fragment_filter or not frag_type)):
        return False
    
    got_flags = Flags.DEFAULT
    if frag_type == 'I':
        got_flags = Flags.CFLAGS_ONLY_I
    else:
        got_flags = Flags.CFLAGS_ONLY_OTHER

    return got_flags in want_flags


@ffi.def_extern()
def filter_libs(client, frag, data) -> bool:
    want_flags, want_fragment_filter = ffi.from_handle(data)

    if Flags.KEEP_SYSTEM_LIBS not in want_flags and lib.pkgconf_fragment_has_system_dir(client, frag):
        return False
    
    frag_type = frag.type.decode()
    if (want_fragment_filter is not None
            and (frag_type not in want_fragment_filter or not frag_type)):
        return False
    
    got_flags = Flags.DEFAULT
    if frag_type == 'L':
        got_flags = Flags.LIBS_ONLY_LDPATH
    elif frag_type == 'l':
        got_flags = Flags.LIBS_ONLY_LIBNAME
    else:
        got_flags = Flags.LIBS_ONLY_OTHER

    return got_flags in want_flags



import pypkgconf

import pytest

import logging
import os
import sys


def test_case_sensitivity(lib1_env):
    assert pypkgconf.query_args("--variable=foo case-sensitivity", env=lib1_env) == [
        "3"
    ]
    assert pypkgconf.query_args("--variable=Foo case-sensitivity", env=lib1_env) == [
        "4"
    ]


def test_depgraph_break_1(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args(
            '--exists --print-errors "foo > 0.6.0 foo < 0.8.0"', env=lib1_env
        )


def test_depgraph_break_2(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args(
            '--exists --print-errors "nonexisting foo <= 3"', env=lib1_env
        )


def test_depgraph_break_3(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--exists --print-errors "depgraph-break"', env=lib1_env)


def test_define_variable(lib1_env):
    result = pypkgconf.query_args(
        '--variable=typelibdir --define-variable="libdir=\\${libdir}" typelibdir',
        env=lib1_env,
    )
    assert result == ["\\${libdir}/typelibdir"]


def test_define_variable_override(lib1_env):
    result = pypkgconf.query_args(
        '--variable=prefix --define-variable="prefix=/test" typelibdir', env=lib1_env
    )
    assert result == ["/test"]


def test_variable(lib1_env):
    result = pypkgconf.query_args("--variable=includedir foo", env=lib1_env)
    assert result == ["/test/include"]


def test_keep_system_libs(lib1_env):
    env = {"LIBRARY_PATH": "/test/local/lib"}
    env.update(lib1_env)

    assert pypkgconf.query_args("--libs-only-L cflags-libs-only", env=env) == []

    result = pypkgconf.query_args(
        "--libs-only-L --keep-system-libs cflags-libs-only", env=env
    )
    assert result == ["-L/test/local/lib"]


def test_libs(lib1_env):
    result = pypkgconf.query_args("--libs cflags-libs-only", env=lib1_env)
    assert result == ["-L/test/local/lib", "-lfoo"]


def test_libs_only(lib1_env):
    result = pypkgconf.query_args(
        "--libs-only-L --libs-only-l cflags-libs-only", env=lib1_env
    )
    assert result == ["-L/test/local/lib", "-lfoo"]


def test_libs_never_mergeback(lib1_env):
    result = pypkgconf.query_args("--libs prefix-foo1", env=lib1_env)
    assert result == ["-L/test/bar/lib", "-lfoo1"]

    result = pypkgconf.query_args("--libs prefix-foo1 prefix-foo2", env=lib1_env)
    assert result == ["-L/test/bar/lib", "-lfoo1", "-lfoo2"]


def test_cflags_only(lib1_env):
    result = pypkgconf.query_args(
        "--cflags-only-I --cflags-only-other cflags-libs-only", env=lib1_env
    )
    assert result == ["-I/test/local/include/foo"]


def test_cflags_never_mergeback(lib1_env):
    result = pypkgconf.query_args("--cflags prefix-foo1 prefix-foo2", env=lib1_env)
    assert result == ["-I/test/bar/include/foo", "-DBAR", "-fPIC", "-DFOO"]


def test_incomplete_libs(lib1_env):
    result = pypkgconf.query_args("--libs incomplete", env=lib1_env)
    assert result == []


def test_incomplete_cflags(lib1_env):
    result = pypkgconf.query_args("--cflags incomplete", env=lib1_env)
    assert result == []


def test_isystem_munge_order(lib1_env):
    result = pypkgconf.query_args("--cflags isystem", env=lib1_env)
    assert result == ["-isystem", "/opt/bad/include", "-isystem", "/opt/bad2/include"]


def test_isystem_munge_sysroot(testsdir, lib1_env):
    env = {"PKG_CONFIG_SYSROOT_DIR": testsdir.as_posix()}
    env.update(lib1_env)

    result = pypkgconf.query_args("--cflags isystem", env=env)

    assert f"-isystem {testsdir.as_posix()}/opt/bad/include" in " ".join(result)


def test_idirafter_munge_order(lib1_env):
    result = pypkgconf.query_args("--cflags idirafter", env=lib1_env)
    assert result == [
        "-idirafter",
        "/opt/bad/include",
        "-idirafter",
        "/opt/bad2/include",
    ]


def test_idirafter_munge_sysroot(testsdir, lib1_env):
    env = {"PKG_CONFIG_SYSROOT_DIR": testsdir.as_posix()}
    env.update(lib1_env)

    result = pypkgconf.query_args("--cflags idirafter", env=env)
    assert f"-idirafter {testsdir.as_posix()}/opt/bad/include" in " ".join(result)


def test_idirafter_ordering(lib1_env):
    result = pypkgconf.query_args("--cflags idirafter-ordering", env=lib1_env)
    assert result == [
        "-I/opt/bad/include1",
        "-idirafter",
        "-I/opt/bad/include2",
        "-I/opt/bad/include3",
    ]


def test_pcpath(testsdir, lib2_env):
    selfdir = testsdir.as_posix()

    result = pypkgconf.query_args(f"--cflags {selfdir}/lib3/bar.pc", env=lib2_env)
    assert result == ["-fPIC", "-I/test/include/foo"]


def test_sysroot_munge(tmp_path, testsdir):
    contents = (testsdir / "lib1" / "sysroot-dir.pc").read_text(encoding="utf-8")
    contents = contents.replace("/sysroot/", testsdir.as_posix() + "/")

    (tmp_path / "lib1").mkdir()
    (tmp_path / "lib1" / "sysroot-dir-selfdir.pc").write_text(
        contents, encoding="utf-8"
    )

    env = {
        "PKG_CONFIG_PATH": (tmp_path / "lib1").as_posix(),
        "PKG_CONFIG_SYSROOT_DIR": testsdir.as_posix(),
    }

    result = pypkgconf.query_args("--libs sysroot-dir-selfdir", env=env)
    assert result == [f"-L{testsdir.as_posix()}/lib", "-lfoo"]


def test_virtual_variable():
    assert pypkgconf.query_args("--exists pkg-config") == []
    assert pypkgconf.query_args("--exists pkgconf") == []

    if sys.platform in {"win32", "cygwin", "msys"}:
        pcpath = "../lib/pkgconfig;../share/pkgconfig"
    else:
        pcpath = os.environ.get("PKG_DEFAULT_PATH")
    result = pypkgconf.query_args("--variable=pc_path pkg-config")
    assert result == [pcpath]

    result = pypkgconf.query_args("--variable=pc_path pkgconf")
    assert result == [pcpath]


def test_fragment_collision(testsdir):
    selfdir = testsdir.as_posix()

    result = pypkgconf.query_args(
        f'--with-path="{selfdir}/lib1" --cflags fragment-collision'
    )
    assert result == ["-D_BAZ", "-D_BAR", "-D_FOO", "-D_THREAD_SAFE", "-pthread"]


def test_malformed_1(testsdir):
    selfdir = testsdir.as_posix()

    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args(f'--validate --with-path="{selfdir}/lib1" malformed-1')


def test_malformed_quoting(testsdir):
    selfdir = testsdir.as_posix()

    assert (
        pypkgconf.query_args(
            f'--validate --with-path="{selfdir}/lib1" malformed-quoting'
        )
        == []
    )


def test_explicit_sysroot(monkeypatch, testsdir):
    # FIXME: does not work with drive letter...
    selfdir = testsdir.as_posix().removeprefix(testsdir.drive)
    monkeypatch.setenv("PKG_CONFIG_SYSROOT_DIR", selfdir)

    result = pypkgconf.query_args(
        f'--debug --with-path="{selfdir}/lib1" --variable=pkgdatadir explicit-sysroot'
    )
    assert result == [f"{selfdir}/usr/share/test"]


def test_empty_tuple(testsdir):
    selfdir = testsdir.as_posix()

    assert (
        pypkgconf.query_args(f'--with-path="{selfdir}/lib1" --cflags empty-tuple') == []
    )


def test_solver_requires_private_debounce(testsdir):
    selfdir = testsdir.as_posix()

    result = pypkgconf.query_args(
        f'--with-path="{selfdir}/lib1" --cflags --libs metapackage'
    )
    assert result == [
        "-I/metapackage-1",
        "-I/metapackage-2",
        "-lmetapackage-1",
        "-lmetapackage-2",
    ]


def test_solver_requires_private_debounce(testsdir, caplog):
    selfdir = testsdir.as_posix()

    with caplog.at_level(logging.WARNING):
        assert (
            pypkgconf.query_args(
                f'--with-path="{selfdir}/lib1" --validate billion-laughs'
            )
            == []
        )

    s = 0
    for r in caplog.records:
        if (
            r.levelno == logging.WARNING
            and r.msg == "warning: truncating very long variable to 64KB"
        ):
            s += 1
    assert s == 5


def test_maximum_package_depth_off_by_one(testsdir):
    selfdir = testsdir.as_posix()

    result = pypkgconf.query_args(
        f'--with-path="{selfdir}/lib1" --modversion foo bar baz'
    )
    assert result == ["1.2.3"]



import pytest

import os
from pathlib import Path
import typing as T


@pytest.fixture
def testsdir() -> Path:
    return Path(__file__).parents[1] / 'subprojects' / 'pkgconf' / 'tests'


@pytest.fixture
def lib1_env(testsdir) -> T.Dict[str, str]:
    return {'PKG_CONFIG_PATH': str(testsdir / 'lib1')}


@pytest.fixture
def lib2_env(testsdir) -> T.Dict[str, str]:
    return {'PKG_CONFIG_PATH': str(testsdir / 'lib2')}


@pytest.fixture
def lib1_lib2_env(lib1_env, lib2_env) -> T.Dict[str, str]:
    paths = [lib1_env['PKG_CONFIG_PATH'], lib2_env['PKG_CONFIG_PATH']]
    return {'PKG_CONFIG_PATH': os.pathsep.join(paths)}



import argparse
from enum import IntFlag
from functools import reduce, wraps
import logging
import operator
import os
import re
import shlex
import sys
import typing as T

from ._libpkgconf import ffi, lib
from .constants import Flags, MAXIMUM_TRAVERSE_DEPTH, OPERATORS
from .logger import logger, set_file_handler, set_stream_handler
from .util import NodeIter

class PkgConfError(RuntimeError): ...


def _render_fragments(fragment_list, fmt='{}') -> T.List[str]:
    result = []

    # Native code does: 
    # render_buf = lib.pkgconf_fragment_render(fragment_list, True, ffi.NULL)
    # result.append(ffi.string(render_buf).decode())
    # but we want to split as list

    for frag in NodeIter(fragment_list, 'pkgconf_fragment_t *'):
        item = ffi.string(frag.data).decode()

        # escape special chars
        if sys.platform not in {'win32', 'cygwin', 'msys'}:
            item = item .replace('\\', '\\\\')
        item = re.sub('[^ $()+-:=@A-Z\\^_a-z~]', R'\\\g<0>', item)
        if not frag.merged:
            item = item.replace(' ', '\\ ')

        if frag.type != b'\0':
            item = f'-{frag.type.decode()}{item}'
        item = fmt.format(item)

        if frag.merged:
            result.extend(item.split(' '))
        else:
            result.append(item)

    return result


def preserve_env(fct):
    @wraps(fct)
    def wrapper(*args, **kwargs):

        orig_env = os.environ.copy()

        try:
            return fct(*args, **kwargs)
        
        finally:
            os.environ.clear()
            os.environ.update(orig_env)

    return wrapper


def relocate_path(path: str) -> str:
    buf = path.encode()
    if len(buf) > lib.PKGCONF_BUFSIZE-1:
        buf = buf[:lib.PKGCONF_BUFSIZE-1]
    cbuf = ffi.new('char[]', buf)

    lib.pkgconf_path_relocate(cbuf, ffi.sizeof(cbuf))
    return ffi.string(cbuf).decode()

def apply_license(client, world, maxdepth) -> T.List[str]:
    result = []
    result_handle = ffi.new_handle(result)
    eflag = lib.pkgconf_pkg_traverse(client, world, lib.print_license, result_handle, maxdepth, 0)
    if eflag != lib.PKGCONF_PKG_ERRF_OK:
        return []
    
    return result

def version() -> str:
    return lib.PACKAGE_VERSION


def apply_uninstalled(client, world, data, maxdepth) -> bool:
    ret_handle = ffi.new_handle(data)
    eflag = lib.pkgconf_pkg_traverse(client, world, lib.check_uninstalled, ret_handle, maxdepth, 0)

    return eflag == lib.PKGCONF_PKG_ERRF_OK


def apply_env_var(prefix, client, world, maxdepth, collect_fn, filter_fn, result, data_handle):
    try:
        unfiltered_list = ffi.new('pkgconf_list_t*')
        filtered_list = ffi.new('pkgconf_list_t*')

        eflag = collect_fn(client, world, unfiltered_list, maxdepth)
        if eflag != lib.PKGCONF_PKG_ERRF_OK:
            return False
        
        lib.pkgconf_fragment_filter(client, filtered_list, unfiltered_list, filter_fn, data_handle)

        if filtered_list.head == ffi.NULL:
            return True
        
        result.extend(_render_fragments(filtered_list, fmt=f"{prefix}='{{}}'"))

    finally:
        lib.pkgconf_fragment_free(unfiltered_list)
        lib.pkgconf_fragment_free(filtered_list)
        

def apply_env(client, world, env_prefix, maxdepth, result, data_handle) -> T.List[str]:
    if not all(c.isalnum() for c in env_prefix):
        return []
    
    result = []

    workbuf = f'{env_prefix}_CFLAGS'
    if not apply_env_var(workbuf, client, world, maxdepth, lib.pkgconf_pkg_cflags, lib.filter_cflags, result, data_handle):
        return []
    
    workbuf = f'{env_prefix}_LIBS'
    if not apply_env_var(workbuf, client, world, maxdepth, lib.pkgconf_pkg_libs, lib.filter_libs, result, data_handle):
        return []
    
    return result


def print_provides(pkg) -> T.List[str]:
    result = []
    for dep in NodeIter(pkg.provides, 'pkgconf_dependency_t *'):
        d = ffi.string(dep.package).decode()
        if dep.version != ffi.NULL:
            comp = ffi.string(lib.pkgconf_pkg_get_comparator(dep)).decode()
            ver = ffi.string(dep.version).decode()
            d += f' {comp} {ver}'
        result.append(d)
    return result


def apply_provides(world) -> T.List[str]:
    result = []
    for dep in NodeIter(world.required, 'pkgconf_dependency_t *'):
        result.extend(print_provides(dep.match))
    return result


def apply_modversion(world) -> T.List[str]:
    result = []
    for dep in NodeIter(world.required, 'pkgconf_dependency_t *'):
        pkg = dep.match
        if pkg.version != ffi.NULL:
            result.append(ffi.string(pkg.version).decode())
    return result


def apply_path(world) -> T.List[str]:
    result = []
    for dep in NodeIter(world.required, 'pkgconf_dependency_t *'):
        pkg = dep.match

        if pkg.filename != ffi.NULL:
            result.append(ffi.string(pkg.filename).decode())
    return result


def print_variables(pkg) -> T.List[str]:
    result = []
    for tuple_ in NodeIter(pkg.vars, 'pkgconf_tuple_t *'):
        result.append(ffi.string(tuple_.key).decode())
    return result


def apply_variables(world) -> T.List[str]:
    result = []
    for dep in NodeIter(world.required, 'pkgconf_dependency_t *'):
        result.extend(print_variables(dep.match))
    return result


def apply_variable(client, world, variable) -> T.List[str]:
    result = []
    for dep in NodeIter(world.required, 'pkgconf_dependency_t *'):
        pkg = dep.match
        var = lib.pkgconf_tuple_find(client, ffi.addressof(pkg.vars), variable.encode())
        if var != ffi.NULL:
            result.append(ffi.string(var).decode())
    return result


def print_requires(pkg) -> T.List[str]:
    result = []
    for dep in NodeIter(pkg.required, 'pkgconf_dependency_t *'):
        d = ffi.string(dep.package).decode()
        if dep.version != ffi.NULL:
            comp = ffi.string(lib.pkgconf_pkg_get_comparator(dep)).decode()
            ver = ffi.string(dep.version).decode()
            d += f' {comp} {ver}'
        result.append(d)
    return result


def apply_requires(world) -> T.List[str]:
    result = []
    for dep in NodeIter(world.required, 'pkgconf_dependency_t *'):
        result.extend(print_requires(dep.match))
    return result


def print_requires_private(pkg) -> T.List[str]:
    result = []
    for dep in NodeIter(pkg.requires_private, 'pkgconf_dependency_t *'):
        d = ffi.string(dep.package).decode()
        if dep.version != ffi.NULL:
            comp = ffi.string(lib.pkgconf_pkg_get_comparator(dep)).decode()
            ver = ffi.string(dep.version).decode()
            d += f' {comp} {ver}'
        result.append(d)
    return result


def apply_requires_private(world) -> T.List[str]:
    result = []
    for dep in NodeIter(world.required, 'pkgconf_dependency_t *'):
        result.extend(print_requires_private(dep.match))
    return result


def apply_cflags(client, world, maxdepth, data) -> T.List[str]:
    try:
        unfiltered_list = ffi.new('pkgconf_list_t*')
        filtered_list = ffi.new('pkgconf_list_t*')

        eflag = lib.pkgconf_pkg_cflags(client, world, unfiltered_list, maxdepth)
        if eflag != lib.PKGCONF_PKG_ERRF_OK:
            return []
        
        lib.pkgconf_fragment_filter(client, filtered_list, unfiltered_list, lib.filter_cflags, data)

        if filtered_list.head == ffi.NULL:
            return []
        
        return _render_fragments(filtered_list)
    
    finally:
        lib.pkgconf_fragment_free(unfiltered_list)
        lib.pkgconf_fragment_free(filtered_list)


def apply_libs(client, world, maxdepth, data) -> T.List[str]:
    try:
        unfiltered_list = ffi.new('pkgconf_list_t*')
        filtered_list = ffi.new('pkgconf_list_t*')

        eflag = lib.pkgconf_pkg_libs(client, world, unfiltered_list, maxdepth)
        if eflag != lib.PKGCONF_PKG_ERRF_OK:
            return []
        
        lib.pkgconf_fragment_filter(client, filtered_list, unfiltered_list, lib.filter_libs, data)

        if filtered_list.head == ffi.NULL:
            return []

        return _render_fragments(filtered_list)
    
    finally:
        lib.pkgconf_fragment_free(unfiltered_list)
        lib.pkgconf_fragment_free(filtered_list)


def parse_args(command: str) -> argparse.Namespace:
    parser = argparse.ArgumentParser(add_help=False, allow_abbrev=False, exit_on_error=False)
    parser.add_argument('--version', dest='want_flags', const=Flags.VERSION | Flags.PRINT_ERRORS, action='append_const')
    # parser.add_argument('--about', dest='want_flags', const=Flags.ABOUT | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--atleast-version', dest='required_module_version')
    parser.add_argument('--atleast-pkgconfig-version', dest='required_pkgconfig_version')
    parser.add_argument('--libs', dest='want_flags', const=Flags.LIBS | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--cflags', dest='want_flags', const=Flags.CFLAGS | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--modversion', dest='want_flags', const=Flags.MODVERSION | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--variable', dest='want_variable')
    parser.add_argument('--exists', dest='want_flags', const=Flags.EXISTS, action='append_const')
    parser.add_argument('--print-errors', dest='want_flags', const=Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--short-errors', dest='want_flags', const=Flags.SHORT_ERRORS, action='append_const')
    parser.add_argument('--maximum-traverse-depth', type=int, default=MAXIMUM_TRAVERSE_DEPTH)
    parser.add_argument('--static', dest='want_flags', const=Flags.STATIC, action='append_const')
    parser.add_argument('--shared', dest='want_flags', const=Flags.SHARED, action='append_const')
    parser.add_argument('--pure', dest='want_flags', const=Flags.PURE, action='append_const')
    parser.add_argument('--print-requires', dest='want_flags', const=Flags.REQUIRES, action='append_const')
    parser.add_argument('--print-variables', dest='want_flags', const=Flags.VARIABLES | Flags.PRINT_ERRORS, action='append_const')
    # parser.add_argument('--digraph', dest='want_flags', const=Flags.DIGRAPH, action='append_const')
    # parser.add_argument('--help', dest='want_flags', const=Flags.HELP, action='append_const')
    parser.add_argument('--env-only', dest='want_flags', const=Flags.ENV_ONLY, action='append_const')
    parser.add_argument('--print-requires-private', dest='want_flags', const=Flags.REQUIRES_PRIVATE, action='append_const')
    parser.add_argument('--cflags-only-I', dest='want_flags', const=Flags.CFLAGS_ONLY_I | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--cflags-only-other', dest='want_flags', const=Flags.CFLAGS_ONLY_OTHER | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--libs-only-L', dest='want_flags', const=Flags.LIBS_ONLY_LDPATH | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--libs-only-l', dest='want_flags', const=Flags.LIBS_ONLY_LIBNAME | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--libs-only-other', dest='want_flags', const=Flags.LIBS_ONLY_OTHER | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--uninstalled', dest='want_flags', const=Flags.UNINSTALLED, action='append_const')
    parser.add_argument('--no-uninstalled', dest='want_flags', const=Flags.NO_UNINSTALLED, action='append_const')
    parser.add_argument('--keep-system-cflags', dest='want_flags', const=Flags.KEEP_SYSTEM_CFLAGS, action='append_const')
    parser.add_argument('--keep-system-libs', dest='want_flags', const=Flags.KEEP_SYSTEM_LIBS, action='append_const')
    parser.add_argument('--define-variable', action='append')
    parser.add_argument('--exact-version', dest='required_exact_module_version')
    parser.add_argument('--max-version', dest='required_max_module_version')
    parser.add_argument('--ignore-conflicts', dest='want_flags', const=Flags.IGNORE_CONFLICTS, action='append_const')
    parser.add_argument('--errors-to-stdout', dest='want_flags', const=Flags.ERRORS_ON_STDOUT, action='append_const')
    parser.add_argument('--silence-errors', dest='want_flags', const=Flags.SILENCE_ERRORS, action='append_const')
    parser.add_argument('--list-all', dest='want_flags', const=Flags.LIST | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--list-package-names', dest='want_flags', const=Flags.LIST_PACKAGE_NAMES | Flags.PRINT_ERRORS, action='append_const')
    # parser.add_argument('--simulate', dest='want_flags', const=Flags.SIMULATE, action='append_const')
    parser.add_argument('--no-cache', dest='want_flags', const=Flags.NO_CACHE, action='append_const')
    parser.add_argument('--print-provides', dest='want_flags', const=Flags.PROVIDES, action='append_const')
    parser.add_argument('--no-provides', dest='want_flags', const=Flags.NO_PROVIDES, action='append_const')
    parser.add_argument('--debug', dest='want_flags', const=Flags.DEBUG | Flags.PRINT_ERRORS, action='append_const')
    parser.add_argument('--validate', dest='want_flags', const=Flags.VALIDATE | Flags.PRINT_ERRORS | Flags.ERRORS_ON_STDOUT, action='append_const')
    parser.add_argument('--log-file', dest='logfile_arg')
    parser.add_argument('--path', dest='want_flags', const=Flags.PATH, action='append_const')
    parser.add_argument('--with-path', action='append')
    parser.add_argument('--prefix-variable')
    parser.add_argument('--define-prefix', dest='want_flags', const=Flags.DEFINE_PREFIX, action='append_const')
    parser.add_argument('--relocate')
    parser.add_argument('--dont-define-prefix', dest='want_flags', const=Flags.DONT_DEFINE_PREFIX, action='append_const')
    parser.add_argument('--dont-relocate-paths', dest='want_flags', const=Flags.DONT_RELOCATE_PATHS, action='append_const')
    parser.add_argument('--env', dest='want_env_prefix')
    # parser.add_argument('--msvc-syntax', dest='want_flags', const=Flags.MSVC_SYNTAX, action='append_const')
    parser.add_argument('--fragment-filter', dest='want_fragment_filter')
    parser.add_argument('--internal-cflags', dest='want_flags', const=Flags.INTERNAL_CFLAGS, action='append_const')
    # parser.add_argument('--dump-personality', dest='want_flags', const=Flags.DUMP_PERSONALITY, action='append_const')
    # parser.add_argument('--personality')
    parser.add_argument('--license', dest='want_flags', const=Flags.DUMP_LICENSE, action='append_const')
    parser.add_argument('args', nargs='*')

    options = parser.parse_args(shlex.split(command))
    options.want_flags = reduce(operator.or_, options.want_flags or [], Flags.DEFAULT)
    if options.define_variable:
        dv = {}
        for opt in options.define_variable:
            key, value = opt.split('=', maxsplit=1)
            dv[key] = value
        options.define_variable = dv

    return options


@preserve_env
def query(args: T.Optional[T.List[str]] = None,
          want_flags: Flags = Flags.DEFAULT,
          env: T.Optional[T.Dict[str, str]] = None,
          required_module_version: T.Optional[str] = None,
          required_pkgconfig_version: T.Optional[str] = None,
          want_variable: T.Optional[str] = None,
          define_variable: T.Optional[T.Dict[str, str]] = None,
          maximum_traverse_depth: int = MAXIMUM_TRAVERSE_DEPTH,
          required_exact_module_version: T.Optional[str] = None,
          required_max_module_version: T.Optional[str] = None,
          logfile_arg: T.Optional[str] = None,
          with_path: T.Optional[T.List[str]] = None,
          prefix_variable: T.Optional[str] = None,
          relocate: T.Optional[str] = None,
          want_env_prefix: T.Optional[str] = None,
          want_fragment_filter: T.Optional[str] = None,
          ) -> T.List[str]:
    
    if env is None:
        env = os.environ.copy()

    else:
        # Those env vars are accessed directly from libpkgconf
        for k in ('PKG_CONFIG_PATH',
                  'PKG_CONFIG_LIBDIR',
                  'PKG_CONFIG_SYSTEM_LIBRARY_PATH',
                  'PKG_CONFIG_SYSTEM_INCLUDE_PATH',
                  'BELIBRARIES',
                  'LIBRARY_PATH',
                  'CPATH',
                  'C_INCLUDE_PATH',
                  'CPLUS_INCLUDE_PATH',
                  'OBJC_INCLUDE_PATH',
                  'INCLUDE'):
            if k in env:
                os.environ[k] = env[k]
            elif k in os.environ:
                del os.environ[k]
    
    if relocate:
        return [relocate_path(relocate)]

    maximum_package_count = 0
    pkgq = ffi.new('pkgconf_list_t*')
    dir_list = ffi.new('pkgconf_list_t*')
    want_client_flags = lib.PKGCONF_PKG_PKGF_NONE

    world = ffi.new('pkgconf_pkg_t *')
    id = ffi.new('char[]', 'virtual:world'.encode())
    world.id = id
    realname = ffi.new('char[]', 'virtual world package'.encode())
    world.realname = realname
    world.flags = lib.PKGCONF_PKG_PROPF_STATIC | lib.PKGCONF_PKG_PROPF_VIRTUAL

    pkg_client = ffi.new('pkgconf_client_t*')

    # if env.get('PKG_CONFIG_EARLY_TRACE'):

    if define_variable:
        for key, value in define_variable.items():
            lib.pkgconf_tuple_define_global(pkg_client, f'{key}={value}'.encode())
    for path in with_path or []:
        lib.pkgconf_path_add(path.encode(), dir_list, True)
    if prefix_variable:
        lib.pkgconf_client_set_prefix_varname(pkg_client, prefix_variable.encode())
        
    personality = lib.pkgconf_cross_personality_default()

    lib.pkgconf_path_copy_list(ffi.addressof(personality.dir_list), dir_list)
    lib.pkgconf_path_free(dir_list)

    error_level_handle = ffi.new_handle(logging.ERROR)
    lib.pkgconf_client_init(pkg_client, lib.error_handler, error_level_handle, personality)

    def cleanup() -> None:
        lib.pkgconf_solution_free(pkg_client, world)
        lib.pkgconf_queue_free(pkgq)
        lib.pkgconf_cross_personality_deinit(personality)
        lib.pkgconf_client_deinit(pkg_client)

        logger.handlers.clear()

    def raise_and_clean(*args) -> None:
        try:
            raise PkgConfError(*args)
        finally:
            cleanup()


    if env.get('PKG_CONFIG_MAXIMUM_TRAVERSE_DEPTH') is not None:
        maximum_traverse_depth = int(env.get('PKG_CONFIG_MAXIMUM_TRAVERSE_DEPTH'))
    
    if Flags.PRINT_ERRORS not in want_flags:
        want_flags |= Flags.SILENCE_ERRORS
    
    if Flags.SILENCE_ERRORS in want_flags and not env.get('PKG_CONFIG_DEBUG_SPEW'):
        want_flags |= Flags.SILENCE_ERRORS
    else:
        want_flags &= ~Flags.SILENCE_ERRORS
    
    if env.get('PKG_CONFIG_DONT_RELOCATE_PATHS'):
        want_flags |= Flags.DONT_RELOCATE_PATHS

    if Flags.VALIDATE in want_flags or Flags.DEBUG in want_flags:
        warning_level_handle = ffi.new_handle(logging.WARNING)
        lib.pkgconf_client_set_warn_handler(pkg_client, lib.error_handler, warning_level_handle)

    #ifndef PKGCONF_LITE
    if Flags.DEBUG in want_flags:
        set_stream_handler(sys.stderr)
        debug_level_handle = ffi.new_handle(logging.DEBUG)
        lib.pkgconf_client_set_trace_handler(pkg_client, lib.error_handler, debug_level_handle)
    #endif

    # if Flags.ABOUT in want_flags

    if Flags.VERSION in want_flags:
        try:
            return [version()]
        finally:
            cleanup()

    # if Flags.HELP in want_flags

    if env.get('PKG_CONFIG_FDO_SYSROOT_RULES'):
        want_client_flags |= lib.PKGCONF_PKG_PKGF_FDO_SYSROOT_RULES

    if env.get('PKG_CONFIG_PKGCONF1_SYSROOT_RULES'):
        want_client_flags |= lib.PKGCONF_PKG_PKGF_PKGCONF1_SYSROOT_RULES

    if Flags.SHORT_ERRORS in want_flags:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_SIMPLIFY_ERRORS
    
    if Flags.DONT_RELOCATE_PATHS in want_flags:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_DONT_RELOCATE_PATHS
    
    if Flags.SILENCE_ERRORS in want_flags:
        set_stream_handler(None)
    elif Flags.ERRORS_ON_STDOUT in want_flags:
        set_stream_handler(sys.stdout)
    else:
        set_stream_handler(sys.stderr)

    if Flags.IGNORE_CONFLICTS in want_flags or env.get('PKG_CONFIG_IGNORE_CONFLICTS'):
        want_client_flags |= lib.PKGCONF_PKG_PKGF_SKIP_CONFLICTS
    
    if Flags.STATIC in want_flags or personality.want_default_static:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_SEARCH_PRIVATE | lib.PKGCONF_PKG_PKGF_MERGE_PRIVATE_FRAGMENTS
    
    if Flags.SHARED in want_flags:
        want_client_flags &= ~(lib.PKGCONF_PKG_PKGF_SEARCH_PRIVATE | lib.PKGCONF_PKG_PKGF_MERGE_PRIVATE_FRAGMENTS)

    if Flags.PURE in want_flags or env.get('PKG_CONFIG_PURE_DEPGRAPH') is not None or personality.want_default_pure:
        want_client_flags &= ~lib.PKGCONF_PKG_PKGF_MERGE_PRIVATE_FRAGMENTS
    
    if Flags.ENV_ONLY in want_flags:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_ENV_ONLY
    
    if Flags.NO_CACHE in want_flags:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_NO_CACHE
    
    # FIXME: only win32?
    if sys.platform in {'win32', 'cygwin', 'msys'} or Flags.DEFINE_PREFIX in want_flags:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_REDEFINE_PREFIX
    
    if Flags.NO_UNINSTALLED in want_flags or env.get('PKG_CONFIG_DISABLE_UNINSTALLED') is not None:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_NO_UNINSTALLED
    
    if Flags.NO_PROVIDES in want_flags:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_SKIP_PROVIDES
    
    if Flags.DONT_DEFINE_PREFIX in want_flags or env.get('PKG_CONFIG_DONT_DEFINE_PREFIX') is not None:
        want_client_flags &= ~lib.PKGCONF_PKG_PKGF_REDEFINE_PREFIX

    if Flags.INTERNAL_CFLAGS in want_flags:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_DONT_FILTER_INTERNAL_CFLAGS
    
    if (Flags.REQUIRES in want_flags
            or Flags.REQUIRES_PRIVATE in want_flags
            or Flags.PROVIDES in want_flags
            or Flags.VARIABLES in want_flags
            or Flags.MODVERSION in want_flags
            or Flags.PATH in want_flags
            or want_variable):
        maximum_package_count = 1
        maximum_traverse_depth = 1
    
    if env.get('PKG_CONFIG_ALLOW_SYSTEM_CFLAGS') is not None:
        want_flags |= Flags.KEEP_SYSTEM_CFLAGS
    
    if env.get('PKG_CONFIG_ALLOW_SYSTEM_LIBS') is not None:
        want_flags |= Flags.KEEP_SYSTEM_LIBS
    
    builddir = env.get('PKG_CONFIG_TOP_BUILD_DIR')
    if builddir is not None:
        lib.pkgconf_client_set_buildroot_dir(pkg_client, builddir.encode())
    
    if Flags.REQUIRES_PRIVATE in want_flags or want_flags & Flags.CFLAGS:
        want_client_flags |= lib.PKGCONF_PKG_PKGF_SEARCH_PRIVATE
    
    sysroot_dir = env.get('PKG_CONFIG_SYSROOT_DIR')
    if sysroot_dir is not None:
        lib.pkgconf_client_set_sysroot_dir(pkg_client, sysroot_dir.encode())

        destdir = env.get('DESTDIR')
        if destdir == sysroot_dir:
            want_client_flags |= lib.PKGCONF_PKG_PKGF_FDO_SYSROOT_RULES
    
    lib.pkgconf_client_set_flags(pkg_client, want_client_flags)
    lib.pkgconf_client_dir_list_build(pkg_client, personality)

    if required_pkgconfig_version is not None:
        if lib.pkgconf_compare_version(lib.PACKAGE_VERSION, required_pkgconfig_version.encode()) >= 0:
            cleanup()
            return []
        raise_and_clean()

    result = []
    result_handle = ffi.new_handle(result)

    if Flags.LIST in want_flags:
        lib.pkgconf_scan_all(pkg_client, result_handle, lib.print_list_entry)
        cleanup()
        return result
    
    if Flags.LIST_PACKAGE_NAMES in want_flags:
        lib.pkgconf_scan_all(pkg_client, result_handle, lib.print_package_entry)
        cleanup()
        return result
    
    if logfile_arg is not None:
        set_file_handler(logfile_arg)

    # Group the 3 version compare, as they are very similar
    module_version_key = None
    module_version_cmp = None

    if required_module_version is not None:
        module_version_key = required_module_version
        module_version_cmp = operator.ge
    elif required_exact_module_version is not None:
        module_version_key = required_exact_module_version
        module_version_cmp = operator.eq
    elif required_max_module_version is not None:
        module_version_key = required_max_module_version
        module_version_cmp = operator.le

    if module_version_key: 
        try:
            deplist = ffi.new('pkgconf_list_t*')

            for arg in args:
                lib.pkgconf_dependency_parse_str(pkg_client, deplist, arg.encode(), 0)

            pkg = ffi.NULL
            for pkgiter in NodeIter(deplist, 'pkgconf_dependency_t *'):
                pkg = lib.pkgconf_pkg_find(pkg_client, pkgiter.package)
                if pkg == ffi.NULL:
                    if Flags.PRINT_ERRORS in want_flags:
                        lib.pkgconf_error(pkg_client, b"Package '%s' was not found\n", pkgiter.package)

                    raise PkgConfError()
                
                if module_version_cmp(lib.pkgconf_compare_version(pkg.version, module_version_key.encode()), 0):
                    return []
                
            raise PkgConfError()

        finally:
            if pkg != ffi.NULL:
                lib.pkgconf_pkg_unref(pkg_client, pkg)
            lib.pkgconf_dependency_free(deplist)
            cleanup()

    while args:
        package = args.pop(0)
        
        if maximum_package_count > 0 and pkgq.length >= maximum_package_count:
            break
        
        # FIXME: probably unneeded
        package.lstrip()
        if not package:
            continue

        if args and args[0] in OPERATORS:
            package = f'{package} {args[0]}, {args[1]}'
            args = args[2:]
        lib.pkgconf_queue_push(pkgq, package.encode())

    if pkgq.head == ffi.NULL:
        raise_and_clean("Please specify at least one package name.")
    
    if not lib.pkgconf_queue_solve(pkg_client, pkgq, world, maximum_traverse_depth):
        raise_and_clean()
    
    if Flags.VALIDATE in want_flags:
        cleanup()
        return []
        
    if Flags.DUMP_LICENSE in want_flags:
        license = apply_license(pkg_client, world, 2)
        cleanup()
        return license

    if Flags.UNINSTALLED in want_flags:
        ret = [False]
        apply_uninstalled(pkg_client, world, ret, 2)
        if ret[0] is False:
            raise_and_clean()

        cleanup()
        return []
    
    if want_env_prefix is not None:
        data = (want_flags, want_fragment_filter)
        data_handle = ffi.new_handle(data)
        result.extend(apply_env(pkg_client, world, want_env_prefix, 2, data_handle))
        want_flags = 0
    
    if Flags.PROVIDES in want_flags:
        want_flags &= ~(Flags.CFLAGS | Flags.LIBS)
        result.extend(apply_provides(world))

    if Flags.MODVERSION in want_flags:
        want_flags &= ~(Flags.CFLAGS | Flags.LIBS)
        result.extend(apply_modversion(world))

    if Flags.PATH in want_flags:
        want_flags &= ~(Flags.CFLAGS | Flags.LIBS)
        lib.pkgconf_client_set_flags(pkg_client, want_client_flags | lib.PKGCONF_PKG_PKGF_SKIP_ROOT_VIRTUAL)
        result.extend(apply_path(world))
    
    if Flags.VARIABLES in want_flags:
        want_flags &= ~(Flags.CFLAGS | Flags.LIBS)
        result.extend(apply_variables(world))

    if want_variable:
        want_flags &= ~(Flags.CFLAGS | Flags.LIBS)
        lib.pkgconf_client_set_flags(pkg_client, want_client_flags | lib.PKGCONF_PKG_PKGF_SKIP_ROOT_VIRTUAL)
        result.extend(apply_variable(pkg_client, world, want_variable))
    
    if Flags.REQUIRES in want_flags:
        want_flags &= ~(Flags.CFLAGS | Flags.LIBS)
        result.extend(apply_requires(world))
    
    if Flags.REQUIRES_PRIVATE in want_flags:
        want_flags &= ~(Flags.CFLAGS | Flags.LIBS)
        result.extend(apply_requires_private(world))
    
    if Flags.CFLAGS & want_flags:
        data = (want_flags, want_fragment_filter)
        data_handle = ffi.new_handle(data)
        result.extend(apply_cflags(pkg_client, world, 2, data_handle))
    
    if Flags.LIBS & want_flags:
        if Flags.STATIC not in want_flags:
            lib.pkgconf_client_set_flags(pkg_client, pkg_client.flags & ~lib.PKGCONF_PKG_PKGF_SEARCH_PRIVATE)

            # redo the solution for the library set: free the solution itself, and any cached graph nodes
            lib.pkgconf_solution_free(pkg_client, world)
            lib.pkgconf_cache_free(pkg_client)

            if not lib.pkgconf_queue_solve(pkg_client, pkgq, world, maximum_traverse_depth):
                raise_and_clean()
        
        data = (want_flags, want_fragment_filter)
        data_handle = ffi.new_handle(data)
        result.extend(apply_libs(pkg_client, world, 2, data_handle))
    
    cleanup()
    return result
        

def query_args(command: str, env: T.Optional[T.Dict[str, str]] = None) -> T.List[str]:
    try:
        options = parse_args(command)
        return query(env=env, **vars(options))
    
    except SystemExit:
        raise PkgConfError()



import pypkgconf


def test_libs(lib1_env):
    result = pypkgconf.query_args('--libs framework-1', env=lib1_env)
    assert result == ['-F/test/lib', '-framework', 'framework-1']

    result = pypkgconf.query_args('--libs framework-2', env=lib1_env)
    assert result == ['-F/test/lib', '-framework', 'framework-2', '-framework', 'framework-1']

    result = pypkgconf.query_args('--libs framework-1 framework-2', env=lib1_env)
    assert result == ['-F/test/lib', '-framework', 'framework-2', '-framework', 'framework-1']



import pypkgconf

import pytest

import sys


skip_pure = pytest.mark.skipif(
    sys.platform in {"win32", "cygwin", "msys"},
    reason="default personality is pure on Windows",
)


def test_libs(lib1_env):
    result = pypkgconf.query_args("--libs bar", env=lib1_env)
    assert result == ["-L/test/lib", "-lbar", "-lfoo"]


def test_cflags(lib1_env):
    result = pypkgconf.query_args("--libs --cflags baz", env=lib1_env)
    assert result == ["-fPIC", "-I/test/include/foo", "-L/test/lib", "-lbaz"]


@skip_pure
def test_libs_static(lib1_env):
    result = pypkgconf.query_args("--static --libs baz", env=lib1_env)
    assert result == [
        "-L/test/lib",
        "-lbaz",
        "-L/test/lib",
        "-lzee",
        "-L/test/lib",
        "-lfoo",
    ]


def test_libs_static_pure(lib1_env):
    result = pypkgconf.query_args("--static --pure --libs baz", env=lib1_env)
    assert result == ["-L/test/lib", "-lbaz", "-L/test/lib", "-lfoo"]


def test_argv_parse2(lib1_env):
    result = pypkgconf.query_args("--static --libs argv-parse-2", env=lib1_env)
    assert result == ["-llib-1", "-pthread", "/test/lib/lib2.so"]


@skip_pure
def test_static_cflags(lib1_env):
    result = pypkgconf.query_args("--static --cflags baz", env=lib1_env)
    assert result == ["-fPIC", "-I/test/include/foo", "-DFOO_STATIC"]


@skip_pure
def test_private_duplication(lib1_env):
    result = pypkgconf.query_args(
        "--static --libs-only-l private-libs-duplication", env=lib1_env
    )
    assert result == ["-lprivate", "-lbaz", "-lzee", "-lbar", "-lfoo", "-lfoo"]


@skip_pure
def test_libs_static2(lib1_env):
    result = pypkgconf.query_args("--static --libs static-libs", env=lib1_env)
    assert result == ["-lbar", "-lbar-private", "-L/test/lib", "-lfoo"]


def test_missing(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args("--cflags missing-require", env=lib1_env)


@skip_pure
def test_requires_internal(testsdir):
    selfdir = testsdir.as_posix()

    result = pypkgconf.query_args(
        f'--with-path="{selfdir}/lib1" --static --libs requires-internal'
    )
    assert result == ["-lbar", "-lbar-private", "-L/test/lib", "-lfoo"]


def test_requires_internal_missing(testsdir):
    selfdir = testsdir.as_posix()

    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args(
            f'--with-path="{selfdir}/lib1" --static --libs requires-internal-missing'
        )


def test_requires_internal_collision(testsdir):
    selfdir = testsdir.as_posix()

    result = pypkgconf.query_args(
        f'--with-path="{selfdir}/lib1" --cflags requires-internal-collision'
    )
    assert result == ["-I/test/local/include/foo"]


def test_orphaned_requires_private(testsdir):
    selfdir = testsdir.as_posix()

    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args(
            f'--with-path="{selfdir}/lib1" --cflags --libs orphaned-requires-private'
        )



from ._libpkgconf import ffi, lib

class NodeIter:

    def __init__(self, pkgconf_list, ctype: str = "pkgconf_tuple_t*"):
        self._cur = pkgconf_list.head
        self._ctype = ctype

    def __iter__(self):
        return self

    def __next__(self):
        if self._cur == ffi.NULL:
            raise StopIteration
        data = self._cur.data
        self._cur = self._cur.next
        return ffi.cast(self._ctype, data)



import pypkgconf

import pytest


def test_simple(lib1_env):
    result = pypkgconf.query_args("--print-provides provides", env=lib1_env)
    assert result == [
        "provides-test-foo = 1.0.0",
        "provides-test-bar > 1.1.0",
        "provides-test-baz >= 1.1.0",
        "provides-test-quux < 1.2.0",
        "provides-test-moo <= 1.2.0",
        "provides-test-meow != 1.3.0",
        "provides = 1.2.3",
    ]

    result = pypkgconf.query_args("--libs provides-request-simple", env=lib1_env)
    assert result == ["-lfoo"]

    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args(
            "--no-provides --libs provides-request-simple", env=lib1_env
        )


def test_foo(lib1_env):
    assert pypkgconf.query_args("--libs provides-test-foo", env=lib1_env) == ["-lfoo"]
    assert pypkgconf.query_args('--libs "provides-test-foo = 1.0.0"', env=lib1_env) == [
        "-lfoo"
    ]
    assert pypkgconf.query_args(
        '--libs "provides-test-foo >= 1.0.0"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-foo <= 1.0.0"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-foo != 1.0.0"', env=lib1_env)
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-foo > 1.0.0"', env=lib1_env)
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-foo < 1.0.0"', env=lib1_env)


def test_bar(lib1_env):
    assert pypkgconf.query_args("--libs provides-test-bar", env=lib1_env) == ["-lfoo"]
    assert pypkgconf.query_args('--libs "provides-test-bar = 1.1.1"', env=lib1_env) == [
        "-lfoo"
    ]
    assert pypkgconf.query_args(
        '--libs "provides-test-bar >= 1.1.1"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-bar <= 1.1.1"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-bar != 1.1.0"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-bar != 1.1.1"', env=lib1_env)
    assert pypkgconf.query_args('--libs "provides-test-bar > 1.1.1"', env=lib1_env) == [
        "-lfoo"
    ]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-bar <= 1.1.0"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-bar <= 1.2.0"', env=lib1_env
    ) == ["-lfoo"]


def test_bar(lib1_env):
    assert pypkgconf.query_args("--libs provides-test-baz", env=lib1_env) == ["-lfoo"]
    assert pypkgconf.query_args('--libs "provides-test-baz = 1.1.0"', env=lib1_env) == [
        "-lfoo"
    ]
    assert pypkgconf.query_args(
        '--libs "provides-test-baz >= 1.1.0"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-baz <= 1.1.0"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-baz != 1.1.0"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-baz !- 1.0.0"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args('--libs "provides-test-baz > 1.1.1"', env=lib1_env) == [
        "-lfoo"
    ]
    assert pypkgconf.query_args('--libs "provides-test-baz > 1.1.0"', env=lib1_env) == [
        "-lfoo"
    ]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-baz < 1.1.0"', env=lib1_env)
    assert pypkgconf.query_args('--libs "provides-test-baz < 1.2.0"', env=lib1_env) == [
        "-lfoo"
    ]


def test_quux(lib1_env):
    assert pypkgconf.query_args("--libs provides-test-quux", env=lib1_env) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-quux = 1.1.9"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-quux >= 1.1.0"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-quux >= 1.1.9"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-quux >= 1.2.0"', env=lib1_env)
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-quux <= 1.2.0"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-quux <= 1.1.9"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-quux != 1.2.0"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-quux != 1.1.0"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-quux > 1.1.9"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-quux > 1.2.0"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-quux < 1.1.0"', env=lib1_env
    ) == ["-lfoo"]
    # FIXME: Test duplicated...
    # with pytest.raises(pypkgconf.PkgConfError):
    #     pypkgconf.query_args('--libs "provides-test-quux > 1.2.0"')


def test_moo(lib1_env):
    assert pypkgconf.query_args("--libs provides-test-moo", env=lib1_env) == ["-lfoo"]
    assert pypkgconf.query_args('--libs "provides-test-moo = 1.2.0"', env=lib1_env) == [
        "-lfoo"
    ]
    assert pypkgconf.query_args(
        '--libs "provides-test-moo >= 1.1.0"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-moo >= 1.2.0"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-moo >= 1.2.1"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-moo <= 1.2.0"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-moo != 1.1.0"', env=lib1_env)
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-moo != 1.0.0"', env=lib1_env)
    assert pypkgconf.query_args('--libs "provides-test-moo > 1.1.9"', env=lib1_env) == [
        "-lfoo"
    ]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-moo > 1.2.0"', env=lib1_env)
    assert pypkgconf.query_args('--libs "provides-test-moo < 1.1.0"', env=lib1_env) == [
        "-lfoo"
    ]
    assert pypkgconf.query_args('--libs "provides-test-moo < 1.2.0"', env=lib1_env) == [
        "-lfoo"
    ]


def test_meow(lib1_env):
    assert pypkgconf.query_args("--libs provides-test-meow", env=lib1_env) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-meow = 1.3.0"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-meow != 1.3.0"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-meow > 1.2.0"', env=lib1_env)
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-meow < 1.3.1"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-meow < 1.3.0"', env=lib1_env
    ) == ["-lfoo"]
    assert pypkgconf.query_args(
        '--libs "provides-test-meow > 1.3.0"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-meow >= 1.3.0"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-meow >= 1.3.1"', env=lib1_env
    ) == ["-lfoo"]
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--libs "provides-test-meow <= 1.3.0"', env=lib1_env)
    assert pypkgconf.query_args(
        '--libs "provides-test-meow < 1.2.0"', env=lib1_env
    ) == ["-lfoo"]


def test_indirect_dependency_node(testsdir):
    selfdir = (testsdir / "lib1").as_posix()

    result = pypkgconf.query_args(
        f'--with-path="{selfdir}" --modversion "provides-test-meow"'
    )
    assert result == ["1.2.3"]

    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args(
            f'--with-path="{selfdir}" --modversion "provides-test-meow = 1.3.0"'
        )



# derived from https://github.com/jeanralphaviles/comment_parser/blob/master/comment_parser/parsers/c_parser.py
# MIT license - https://github.com/jeanralphaviles/comment_parser/blob/master/LICENSE


class Comment:
    def __init__(self, comment, line, multiline):
        self.comment = comment
        self.line = line
        self.multiline = multiline

    def __repr__(self):
        return "Comment(comment=%r, line=%r, multiline=%r)" % (self.comment, self.line, self.multiline)

    @property
    def clean_text(self):
        if not self.multiline:
            return self.comment.strip()

        lines = self.comment.splitlines()
        cleanlines = []
        for line in lines:
            if line[0:3] == ' * ':
                cleanlines.append(line[3:])
            elif len(line) == 2:
                cleanlines.append('')
        return '\n'.join(cleanlines)

    @property
    def doc_text(self):
        text = self.clean_text
        if '!doc' in text[0:4]:
            return text[5:]
        return None


class FileError(Exception):
    pass


class UnterminatedCommentError(Exception):
    pass


def extract_comments(filename):
    """Extracts a list of comments from the given C family source file.
    Comments are represented with the Comment class found in the common module.
    C family comments come in two forms, single and multi-line comments.
        - Single-line comments begin with '//' and continue to the end of line.
        - Multi-line comments begin with '/*' and end with '*/' and can span
            multiple lines of code. If a multi-line comment does not terminate
            before EOF is reached, then an exception is raised.
    Note that this doesn't take language-specific preprocessor directives into
    consideration.
    Args:
        filename: String name of the file to extract comments from.
    Returns:
        Python list of Comment objects in the order that they appear in the file.
    Raises:
        FileError: File was unable to be open or read.
        UnterminatedCommentError: Encountered an unterminated multi-line
            comment.
    """
    try:
        with open(filename, 'r') as source_file:
            state = 0
            current_comment = ''
            comments = []
            line_counter = 1
            comment_start = 1
            while True:
                char = source_file.read(1)
                if not char:
                    if state == 3 or state == 4:
                        raise UnterminatedCommentError()
                    if state == 2:
                        # Was in single line comment. Create comment.
                        comment = Comment(current_comment, line_counter, False)
                        comments.append(comment)
                    return comments
                if state == 0:
                    # Waiting for comment start character or beginning of
                    # string.
                    if char == '/':
                        state = 1
                    elif char == '"':
                        state = 5
                elif state == 1:
                    # Found comment start character, classify next character and
                    # determine if single or multiline comment.
                    if char == '/':
                        state = 2
                    elif char == '*':
                        comment_start = line_counter
                        state = 3
                    else:
                        state = 0
                elif state == 2:
                    # In single line comment, read characters until EOL.
                    if char == '\n':
                        comment = Comment(current_comment, line_counter, False)
                        comments.append(comment)
                        current_comment = ''
                        state = 0
                    else:
                        current_comment += char
                elif state == 3:
                    # In multi-line comment, add characters until '*'
                    # encountered.
                    if char == '*':
                        state = 4
                    else:
                        current_comment += char
                elif state == 4:
                    # In multi-line comment with asterisk found. Determine if
                    # comment is ending.
                    if char == '/':
                        comment = Comment(
                            current_comment, comment_start, True)
                        comments.append(comment)
                        current_comment = ''
                        state = 0
                    else:
                        current_comment += '*'
                        # Care for multiple '*' in a row
                        if char != '*':
                            current_comment += char
                            state = 3
                elif state == 5:
                    # In string literal, expect literal end or escape char.
                    if char == '"':
                        state = 0
                    elif char == '\\':
                        state = 6
                elif state == 6:
                    # In string literal, escaping current char.
                    state = 5
                if char == '\n':
                    line_counter += 1
    except OSError as exception:
        raise FileError(str(exception))


if __name__ == '__main__':
    import sys
    from pprint import pprint

    comments = [comment for comment in extract_comments(sys.argv[1]) if comment.doc_text]
    for comment in comments:
        print(comment.doc_text)



import pypkgconf

import pytest

import logging


def test_comments(lib1_env):
    result = pypkgconf.query_args("--libs comments", env=lib1_env)

    assert result == ["-lfoo"]


def test_comments_in_fields(lib1_env):
    result = pypkgconf.query_args("--libs comments-in-fields", env=lib1_env)

    assert result == ["-lfoo"]


def test_dos(lib1_env):
    result = pypkgconf.query_args("--libs dos-lineendings", env=lib1_env)

    assert result == ["-L/test/lib/dos-lineendings", "-ldos-lineendings"]


def test_no_trailing_newline(lib1_env):
    result = pypkgconf.query_args("--cflags no-trailing-newline", env=lib1_env)

    assert result == ["-I/test/include/no-trailing-newline"]


def test_parse(lib1_env):
    result = pypkgconf.query_args("--libs argv-parse", env=lib1_env)

    assert result == ["-llib-3", "-llib-1", "-llib-2", "-lpthread"]


def test_bad_option(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args("--exists -foo", env=lib1_env)


def test_argv_parse_3(lib1_env):
    result = pypkgconf.query_args("--libs argv-parse-3", env=lib1_env)

    assert result == ["-llib-1", "-pthread", "/test/lib/lib2.so"]


def test_quoting(lib1_env):
    result = pypkgconf.query_args("--libs tilde-quoting", env=lib1_env)
    assert result == ["-L~", "-ltilde"]

    result = pypkgconf.query_args("--cflags tilde-quoting", env=lib1_env)
    assert result == ["-I~"]


def test_paren_quoting(lib1_env):
    result = pypkgconf.query_args("--libs paren-quoting", env=lib1_env)

    # FIXME: escape $?
    assert result == ["-L$(libdir)", "-ltilde"]


def test_multiline_field(lib1_env):
    result = pypkgconf.query_args("--list-all", env=lib1_env)

    assert "multiline description" in "\n".join(result)


def test_multiline_bogus_header(lib1_env):
    assert pypkgconf.query_args("--exists multiline-bogus", env=lib1_env) == []


def test_escaped_backslash(testsdir):
    selfdir = (testsdir / "lib1").as_posix()
    result = pypkgconf.query_args(f"--with-path={selfdir} --cflags escaped-backslash")

    assert result == ["-IC:\\\\A"]


def test_quoted(lib1_env):
    result = pypkgconf.query_args("--cflags quotes", env=lib1_env)

    # FIXME: is it ok?
    assert result == [
        '-DQUOTED=\\"bla\\"',
        '-DA=\\"escaped\\ string\\\'\\ literal\\"',
        "-DB=\\\\1$",
        "-DC=bla",
    ]


def test_flag_order_1(lib1_env):
    result = pypkgconf.query_args("--libs flag-order-1", env=lib1_env)

    assert result == ["-L/test/lib", "-Bdynamic", "-lfoo", "-Bstatic", "-lbar"]


def test_flag_order_2(lib1_env):
    result = pypkgconf.query_args("--libs flag-order-1 foo", env=lib1_env)

    assert result == ["-L/test/lib", "-Bdynamic", "-lfoo", "-Bstatic", "-lbar", "-lfoo"]


def test_flag_order_3(lib1_env):
    result = pypkgconf.query_args("--libs flag-order-3", env=lib1_env)

    assert result == [
        "-L/test/lib",
        "-Wl,--start-group",
        "-lfoo",
        "-lbar",
        "-Wl,--end-group",
    ]


def test_flag_order_4(lib1_env):
    result = pypkgconf.query_args("--libs flag-order-3 foo", env=lib1_env)

    assert result == [
        "-L/test/lib",
        "-Wl,--start-group",
        "-lfoo",
        "-lbar",
        "-Wl,--end-group",
        "-lfoo",
    ]


def test_variable_whitespace(lib1_env):
    result = pypkgconf.query_args("--cflags variable-whitespace", env=lib1_env)

    assert result == ["-I/test/include"]


def test_fragment_quoting(lib1_env):
    result = pypkgconf.query_args("--cflags fragment-quoting", env=lib1_env)

    assert result == ["-fPIC", "-I/test/include/foo", '-DQUOTED=\\"/test/share/doc\\"']


def test_fragment_quoting_2(lib1_env):
    result = pypkgconf.query_args("--cflags fragment-quoting-2", env=lib1_env)

    assert result == ["-fPIC", "-I/test/include/foo", "-DQUOTED=/test/share/doc"]


def test_fragment_quoting_3(lib1_env):
    result = pypkgconf.query_args("--cflags fragment-quoting-3", env=lib1_env)

    assert result == ["-fPIC", "-I/test/include/foo", '-DQUOTED=\\"/test/share/doc\\"']


def test_fragment_quoting_5(lib1_env):
    result = pypkgconf.query_args("--cflags fragment-quoting-5", env=lib1_env)

    assert result == ["-fPIC", "-I/test/include/foo", "-DQUOTED=/test/share/doc"]


def test_fragment_quoting_7(lib1_env):
    result = pypkgconf.query_args("--cflags fragment-quoting-7", env=lib1_env)

    assert result == [
        "-Dhello=10",
        "-Dworld=+32",
        "-DDEFINED_FROM_PKG_CONFIG=hello\\ world",
    ]


def test_fragment_escaping_1(testsdir):
    selfdir = (testsdir / "lib1").as_posix()
    result = pypkgconf.query_args(f"--with-path={selfdir} --cflags fragment-escaping-1")

    assert result == ["-IC:\\\\D\\ E"]


def test_fragment_escaping_2(testsdir):
    selfdir = (testsdir / "lib1").as_posix()
    result = pypkgconf.query_args(f"--with-path={selfdir} --cflags fragment-escaping-2")

    assert result == ["-IC:\\\\D\\ E"]


def test_fragment_escaping_3(testsdir):
    selfdir = (testsdir / "lib1").as_posix()
    result = pypkgconf.query_args(f"--with-path={selfdir} --cflags fragment-escaping-3")

    assert result == ["-IC:\\\\D\\ E"]


# FIXME: how to test macro arithmetics in Python...
# def test_fragment_quoting_7a(testsdir):
#     selfdir = (testsdir / 'lib1').as_posix()
#     result = pypkgconf.query_args(f'--with-path={selfdir} --cflags fragment-quoting-7')

#     assert result == ['-Dhello=10', '-Dworld=+32', '-DDEFINED_FROM_PKG_CONFIG=hello\\ world']

# 	cat > test.c <<- __TESTCASE_END__
# 		int main(int argc, char *argv[]) { return DEFINED_FROM_PKG_CONFIG; }
# 	__TESTCASE_END__
# 	cc -o test-fragment-quoting-7 ${test_cflags} ./test.c
# 	atf_check -e 42 ./test-fragment-quoting-7
# 	rm -f test.c test-fragment-quoting-7


def test_fragment_comment(testsdir):
    selfdir = (testsdir / "lib1").as_posix()
    result = pypkgconf.query_args(f"--with-path={selfdir} --cflags fragment-comment")

    assert result == ["kuku=\\#ttt"]


@pytest.mark.skip(reason="--msvc-syntax not implemented")
def test_msvc_fragment_quoting(lib1_env):
    result = pypkgconf.query_args(
        "--libs --msvc-syntax fragment-escaping-1", env=lib1_env
    )

    assert result == ['/libpath:"C:\\D E"', "E.lib"]


@pytest.mark.skip(reason="--msvc-syntax not implemented")
def test_msvc_fragment_render_cflags(lib1_env):
    result = pypkgconf.query_args("--cflags --static --msvc-syntax foo", env=lib1_env)

    assert result == ["/I/test/include/foo", "/DFOO_STATIC"]


def test_tuple_dequote(testsdir):
    selfdir = (testsdir / "lib1").as_posix()
    result = pypkgconf.query_args(f"--with-path={selfdir} --libs tuple-quoting")

    assert result == ["-L/test/lib", "-lfoo"]


def test_version_with_whitespace(testsdir):
    selfdir = (testsdir / "lib1").as_posix()
    result = pypkgconf.query_args(
        f"--with-path={selfdir} --modversion malformed-version"
    )

    assert result == ["3.922"]


def test_version_with_whitespace_2(testsdir):
    selfdir = (testsdir / "lib1").as_posix()
    result = pypkgconf.query_args(
        f"--with-path={selfdir} --print-provides malformed-version"
    )

    assert result == ["malformed-version = 3.922"]


def test_version_with_whitespace_diagnostic(testsdir, caplog):
    selfdir = (testsdir / "lib1").as_posix()

    with caplog.at_level(logging.WARNING):
        result = pypkgconf.query_args(
            f"--with-path={selfdir} --validate malformed-version"
        )

    assert result == []
    assert any(record.levelname == "WARNING" for record in caplog.records)



import pypkgconf

import pytest

import shutil


def test_noargs(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args("", env=lib1_env)


def test_libs(lib1_env):
    result = pypkgconf.query_args("--libs foo", env=lib1_env)

    assert result == ["-L/test/lib", "-lfoo"]


def test_libs_cflags(lib1_env):
    result = pypkgconf.query_args("--cflags --libs foo", env=lib1_env)

    assert result == ["-fPIC", "-I/test/include/foo", "-L/test/lib", "-lfoo"]


def test_libs_cflags_version(lib1_env):
    result = pypkgconf.query_args('--cflags --libs "foo > 1.2"', env=lib1_env)

    assert result == ["-fPIC", "-I/test/include/foo", "-L/test/lib", "-lfoo"]


def test_libs_cflags_version_multiple(lib1_env):
    result = pypkgconf.query_args(
        '--cflags --libs "foo > 1.2 bar >= 1.3"', env=lib1_env
    )

    assert result == ["-fPIC", "-I/test/include/foo", "-L/test/lib", "-lbar", "-lfoo"]


def test_libs_cflags_version_multiple_coma(lib1_env):
    result = pypkgconf.query_args(
        '--cflags --libs "foo > 1.2,bar >= 1.3"', env=lib1_env
    )

    assert result == ["-fPIC", "-I/test/include/foo", "-L/test/lib", "-lbar", "-lfoo"]


def test_libs_cflags_version_alt(lib1_env):
    result = pypkgconf.query_args('--cflags --libs "foo" ">" "1.2"', env=lib1_env)

    assert result == ["-fPIC", "-I/test/include/foo", "-L/test/lib", "-lfoo"]


def test_libs_cflags_version_different(lib1_env):
    result = pypkgconf.query_args('--cflags --libs "foo" "!=" "1.3.0"', env=lib1_env)

    assert result == ["-fPIC", "-I/test/include/foo", "-L/test/lib", "-lfoo"]


def test_libs_cflags_version_different_bad(caplog, lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--cflags --libs "foo" "!=" "1.2.3"', env=lib1_env)

    assert (
        "Package dependency requirement 'foo != 1.2.3' could not be satisfied.\n"
        in caplog.text
    )
    assert (
        "Package 'foo' has version '1.2.3', required version is '!= 1.2.3'\n"
        in caplog.text
    )


def test_exists_nonexitent(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args("--exists nonexistant", env=lib1_env)


def test_nonexitent(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args("nonexistant", env=lib1_env)


def test_exists_version(lib1_env):
    assert pypkgconf.query_args('--exists "foo > 1.2"', env=lib1_env) == []


def test_exists_version_bad(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--exists "foo > 1.2.3"', env=lib1_env)


def test_exists_version_bad(lib1_env):
    assert pypkgconf.query_args('--exists "foo" ">" "1.2"', env=lib1_env) == []


def test_uninstalled_bad(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--uninstalled "foo"', env=lib1_env)


def test_uninstalled(lib1_env):
    assert pypkgconf.query_args('--uninstalled "omg"', env=lib1_env) == []


def test_exists_version_bad2(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--exists "foo >= "', env=lib1_env)


def test_exists_version_bad3(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args('--exists "tilde >= 1.0.0"', env=lib1_env)


def test_exists(lib1_env):
    assert pypkgconf.query_args('--exists "tilde = 1.0.0~rc1"', env=lib1_env) == []


def test_exists2(lib1_env):
    assert pypkgconf.query_args('--exists "tilde <= 1.0.0"', env=lib1_env) == []


def test_exists3(lib1_env):
    assert pypkgconf.query_args('--exists "" "foo"', env=lib1_env) == []


def test_libs_intermediary(lib1_env):
    result = pypkgconf.query_args("--libs intermediary-1 intermediary-2", env=lib1_env)

    assert result == ["-lintermediary-1", "-lintermediary-2", "-lfoo", "-lbar", "-lbaz"]


def test_libs_circular2(caplog, lib1_env):
    assert pypkgconf.query_args("circular-2 --validate", env=lib1_env) == []
    assert (
        "circular-1: breaking circular reference (circular-1 -> circular-2 -> circular-1)\n"
        in caplog.text
    )


def test_libs_circular1(caplog, lib1_env):
    assert pypkgconf.query_args("circular-1 --validate", env=lib1_env) == []
    assert (
        "circular-3: breaking circular reference (circular-3 -> circular-1 -> circular-3)\n"
        in caplog.text
    )


def test_libs_circular_directpc(testsdir, lib1_env):
    result = pypkgconf.query_args(
        f"--libs {testsdir.as_posix()}/lib1/circular-3.pc", env=lib1_env
    )

    assert result == ["-lcircular-2", "-lcircular-3", "-lcircular-1"]


def test_libs_static(lib1_env):
    result = pypkgconf.query_args("--libs static-archive-libs", env=lib1_env)

    assert result == ["/libfoo.a", "-pthread"]


def test_libs_static_ordering(lib1_env):
    result = pypkgconf.query_args("--libs foo bar", env=lib1_env)

    assert result == ["-L/test/lib", "-lbar", "-lfoo"]


def test_pkg_config_path(lib1_lib2_env):
    result = pypkgconf.query_args("--libs foo", env=lib1_lib2_env)

    assert result == ["-L/test/lib", "-lfoo"]

    result = pypkgconf.query_args("--libs bar", env=lib1_lib2_env)

    assert result == ["-L/test/lib", "-lbar", "-lfoo"]


def test_with_path(testsdir):
    lib1_path = testsdir / "lib1"
    lib2_path = testsdir / "lib2"

    result = pypkgconf.query_args(
        f"--with-path={lib1_path.as_posix()} --with-path={lib2_path.as_posix()} --libs foo"
    )

    assert result == ["-L/test/lib", "-lfoo"]

    result = pypkgconf.query_args(
        f"--with-path={lib1_path.as_posix()} --with-path={lib2_path.as_posix()} --libs bar"
    )

    assert result == ["-L/test/lib", "-lbar", "-lfoo"]


def test_nolibs(lib1_env):
    assert pypkgconf.query_args("--libs nolib", env=lib1_env) == []


def test_nocflags(lib1_env):
    assert pypkgconf.query_args("--cflags nocflag", env=lib1_env) == []


def test_arbitary_path(testsdir, tmp_path, monkeypatch):
    shutil.copy(testsdir / "lib1" / "foo.pc", tmp_path)
    monkeypatch.chdir(tmp_path)

    result = pypkgconf.query_args("--libs foo.pc")

    assert result == ["-L/test/lib", "-lfoo"]


def test_relocatable(testsdir):
    basedir = pypkgconf.query_args(f"--relocate {testsdir.as_posix()}")[0]

    result = pypkgconf.query_args(
        f"--define-prefix --variable=prefix {basedir}/lib-relocatable/lib/pkgconfig/foo.pc"
    )

    assert result == [f"{basedir}/lib-relocatable"]


def test_single_depth_selectors(testsdir):
    env = {"PKG_CONFIG_MAXIMUM_TRAVERSE_DEPTH": "1"}

    result = pypkgconf.query_args(
        f"--with-path={testsdir.as_posix()}/lib3 --print-requires bar", env=env
    )

    assert result == ["foo"]


def test_license_isc(testsdir):
    result = pypkgconf.query_args(
        f"--with-path={testsdir.as_posix()}/lib1 --license foo"
    )

    assert result == ["foo: ISC"]


def test_license_noassertion(testsdir):
    result = pypkgconf.query_args(
        f"--with-path={testsdir.as_posix()}/lib1 --license bar"
    )

    assert result == ["bar: NOASSERTION", "foo: ISC"]


def test_modversion_noflatten(testsdir):
    result = pypkgconf.query_args(
        f"--with-path={testsdir.as_posix()}/lib1 --modversion bar"
    )

    assert result == ["1.3"]



import pypkgconf


def test_libs(lib1_env):
    result = pypkgconf.query_args('--libs conflicts', env=lib1_env)

    assert result == ['-L/test/lib', '-lconflicts']


def test_ignore(lib1_env):
    result = pypkgconf.query_args('--ignore-conflicts --libs conflicts', env=lib1_env)

    assert result == ['-L/test/lib', '-lconflicts']



from cffi import FFI

from pathlib import Path
import re
import sys


def sanitize_lipkgconf_h(lipkgconf_h: Path) -> str:
    cdef = lipkgconf_h.read_text()
    cdef = cdef.replace('#ifndef LIBPKGCONF__LIBPKGCONF_H\n', '')
    cdef = cdef.replace('#define LIBPKGCONF__LIBPKGCONF_H\n', '')
    cdef = re.sub('#include <.+>\n', '', cdef)
    cdef = re.sub('#if ((.|\n)*?)#endif.*\n', '', cdef, 0)
    cdef = re.sub('#ifn?def ((.|\n)*?)#endif.*\n', '', cdef, 0)
    cdef = re.sub('#define.*["()](.*\\\\\n)*(.*\n)', '', cdef)
    cdef = cdef.replace('#endif\n', '')
    cdef = cdef.replace('PKGCONF_API ', '')
    cdef = re.sub(r' PRINTFLIKE\(.*?\)', '', cdef)
    return cdef


if __name__ == '__main__':
    lipkgconf_h = Path(sys.argv[1], 'libpkgconf', 'libpkgconf.h')
    cdef = """
// from iter.h
typedef struct pkgconf_node_ pkgconf_node_t;

struct pkgconf_node_ {
	pkgconf_node_t *prev, *next;
	void *data;
};

typedef struct {
	pkgconf_node_t *head, *tail;
	size_t length;
} pkgconf_list_t;
"""
    cdef += sanitize_lipkgconf_h(lipkgconf_h)
    cdef += """
#define PKGCONF_BUFSIZE ...

/* Python callbacks */
extern "Python" bool error_handler(const char *msg, const pkgconf_client_t *client, void *data);
extern "Python" bool print_list_entry(const pkgconf_pkg_t *entry, void *data);
extern "Python" bool print_package_entry(const pkgconf_pkg_t *entry, void *data);
extern "Python" void print_license(pkgconf_client_t *client, pkgconf_pkg_t *pkg, void *data);
extern "Python" void check_uninstalled(pkgconf_client_t *client, pkgconf_pkg_t *pkg, void *data);
extern "Python" bool filter_cflags(const pkgconf_client_t *client, const pkgconf_fragment_t *frag, void *data);
extern "Python" bool filter_libs(const pkgconf_client_t *client, const pkgconf_fragment_t *frag, void *data);
"""

    tmp = Path(sys.argv[2]).with_suffix('.cdef')
    tmp.write_text(cdef)

    ffibuilder = FFI()
    ffibuilder.set_source('_libpkgconf', '#include <libpkgconf/libpkgconf.h>')
    ffibuilder.cdef(cdef)

    ffibuilder.emit_c_code(sys.argv[2])



import pypkgconf

import pytest


def test_atleast(lib1_env):
    assert pypkgconf.query_args("--atleast-version 1.0 foo", env=lib1_env) == []

    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args("--atleast-version 2.0 foo", env=lib1_env)


def test_exact(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args("--exact-version 1.0 foo", env=lib1_env)

    assert pypkgconf.query_args("--exact-version 1.2.3 foo", env=lib1_env) == []


def test_max(lib1_env):
    with pytest.raises(pypkgconf.PkgConfError):
        pypkgconf.query_args("--max-version 1.0 foo", env=lib1_env)

    assert pypkgconf.query_args("--max-version 2.0 foo", env=lib1_env) == []



from enum import IntFlag

class Flags(IntFlag):
    DEFAULT = 0

    CFLAGS_ONLY_I = (1 << 2)
    CFLAGS_ONLY_OTHER = (1 << 3)
    CFLAGS = (CFLAGS_ONLY_I|CFLAGS_ONLY_OTHER)
    LIBS_ONLY_LDPATH = (1 << 5)
    LIBS_ONLY_LIBNAME = (1 << 6)
    LIBS_ONLY_OTHER = (1 << 7)
    LIBS = (LIBS_ONLY_LDPATH|LIBS_ONLY_LIBNAME|LIBS_ONLY_OTHER)
    MODVERSION = (1 << 8)
    REQUIRES = (1 << 9)
    REQUIRES_PRIVATE = (1 << 10)
    VARIABLES = (1 << 11)
    DIGRAPH = (1 << 12)
    KEEP_SYSTEM_CFLAGS = (1 << 13)
    KEEP_SYSTEM_LIBS = (1 << 14)
    VERSION = (1 << 15)
    ABOUT = (1 << 16)
    ENV_ONLY = (1 << 17)
    ERRORS_ON_STDOUT = (1 << 18)
    SILENCE_ERRORS = (1 << 19)
    IGNORE_CONFLICTS = (1 << 20)
    STATIC = (1 << 21)
    NO_UNINSTALLED = (1 << 22)
    UNINSTALLED = (1 << 23)
    LIST = (1 << 24)
    HELP = (1 << 25)
    PRINT_ERRORS = (1 << 26)
    SIMULATE = (1 << 27)
    NO_CACHE = (1 << 28)
    PROVIDES = (1 << 29)
    VALIDATE = (1 << 30)
    LIST_PACKAGE_NAMES = (1 << 31)
    NO_PROVIDES = (1 << 32)
    PURE = (1 << 33)
    PATH = (1 << 34)
    DEFINE_PREFIX = (1 << 35)
    DONT_DEFINE_PREFIX = (1 << 36)
    DONT_RELOCATE_PATHS = (1 << 37)
    DEBUG = (1 << 38)
    SHORT_ERRORS = (1 << 39)
    EXISTS = (1 << 40)
    MSVC_SYNTAX = (1 << 41)
    INTERNAL_CFLAGS = (1 << 42)
    DUMP_PERSONALITY = (1 << 43)
    SHARED = (1 << 44)
    DUMP_LICENSE = (1 << 45)

MAXIMUM_TRAVERSE_DEPTH = 2000

OPERATORS = {'<', '>', '<=', '>=', '==', '!='}



import logging
import typing as T

logger = logging.getLogger('pkgconf')
formatter = logging.Formatter('%(asctime)s %(levelname)s:%(name)s: %(message)s')


def set_stream_handler(stream):
    logger.handlers.clear()

    handler = logging.StreamHandler(stream)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def set_file_handler(file):
    logger.handlers.clear()

    handler = logging.FileHandler(file)
    handler.setFormatter(formatter)
    logger.addHandler(handler)



import pypkgconf


def test_do_not_eat_slash(lib1_env):
    env = {"PKG_CONFIG_SYSROOT_DIR": "/"}
    env.update(lib1_env)

    result = pypkgconf.query_args("--cflags baz", env=env)
    assert result == ["-fPIC", "-I/test/include/foo"]


def test_cflags(testsdir, lib1_env):
    selfdir = testsdir.as_posix()
    env = {"PKG_CONFIG_SYSROOT_DIR": selfdir}
    env.update(lib1_env)

    result = pypkgconf.query_args("--cflags baz", env=env)
    assert result == ["-fPIC", f"-I{selfdir}/test/include/foo"]


def test_variable(testsdir, lib1_env):
    selfdir = testsdir.as_posix()
    env = {"PKG_CONFIG_SYSROOT_DIR": selfdir}
    env.update(lib1_env)

    result = pypkgconf.query_args("--variable=prefix foo", env=env)
    assert result == [f"{selfdir}/test"]

    result = pypkgconf.query_args("--variable=includedir foo", env=env)
    assert result == [f"{selfdir}/test/include"]


def test_do_not_duplicate_sysroot_dir(testsdir, lib1_env):
    env = {"PKG_CONFIG_SYSROOT_DIR": "/sysroot"}
    env.update(lib1_env)

    result = pypkgconf.query_args("--cflags sysroot-dir-2", env=env)
    assert result == ["-I/sysroot/usr/include"]

    result = pypkgconf.query_args("--cflags sysroot-dir-3", env=env)
    assert result == ["-I/sysroot/usr/include"]

    result = pypkgconf.query_args("--cflags sysroot-dir-5", env=env)
    assert result == ["-I/sysroot/usr/include"]

    selfdir = testsdir.as_posix()
    env["PKG_CONFIG_SYSROOT_DIR"] = selfdir
    result = pypkgconf.query_args("--cflags sysroot-dir-4", env=env)
    assert result == [f"-I{selfdir}/usr/include"]


def test_uninstalled(lib1_env):
    env = {"PKG_CONFIG_SYSROOT_DIR": "/sysroot"}
    env.update(lib1_env)

    result = pypkgconf.query_args("--libs omg", env=env)
    assert result == ["-L/test/lib", "-lomg"]


def test_uninstalled_pkgconf1(lib1_env):
    env = {
        "PKG_CONFIG_SYSROOT_DIR": "/sysroot",
        "PKG_CONFIG_PKGCONF1_SYSROOT_RULES": "1",
    }
    env.update(lib1_env)

    result = pypkgconf.query_args("--libs omg", env=env)
    assert result == ["-L/sysroot/test/lib", "-lomg"]


def test_uninstalled_fdo(lib1_env):
    env = {
        "PKG_CONFIG_SYSROOT_DIR": "/sysroot",
        "PKG_CONFIG_FDO_SYSROOT_RULES": "1",
    }
    env.update(lib1_env)

    result = pypkgconf.query_args("--libs omg", env=env)
    assert result == ["-L/test/lib", "-lomg"]


def test_uninstalled_fdo_pc_sysrootdir(lib1_env):
    env = {
        "PKG_CONFIG_SYSROOT_DIR": "/sysroot",
        "PKG_CONFIG_FDO_SYSROOT_RULES": "1",
    }
    env.update(lib1_env)

    result = pypkgconf.query_args("--libs omg-sysroot", env=env)
    assert result == ["-L/sysroot/test/lib", "-lomg"]
